
class PaymentService {
    constructor(opts, imports) {
        if (typeof this.paymentStatus != 'object' || (typeof this.serialPort != 'object' &&
            typeof this.netSocket  != 'object')){
            this.paymentStatus = {};
        }
    }
    
    /*
    *
    *
    *   Return the progression of the payment step (Init, Progress, Success, Error, ...)
    *
    */
    getPaymentState(postParam){
        console.log(this.paymentStatus, postParam.token);
        return this.paymentStatus[postParam.token];
    }
    /**
     *
     *
     * Charger la liste des notes dans le cache.
     *
     *
     *
     */
    makePayment(postParam) {
        var self = this; 
        return new Promise((resolve) => {
                /*
                * 
                *   Dave Hill | dave-hill@dyrk.org
                *   USING PROTOCOL CONCERT
                *   ref : https://lists.launchpad.net/openerp-community/pdfcezlBjgtdJ.pdf
                */
               let  SerialPort  = require('serialport'),
                    net   = require('net');
                /*
                *
                *   Socket : Init Communication by Ethernet / Wireless
                *   Take in params the value to send on TPE
                *   & an "Array" (Object) with the expected value binded with their function to run
                * 
                */
               var TPEByEthernet = function(req, listAction){
                    console.log('TPE => Net.Socket');
                    self.config.listAction = listAction;
                    self.config.lastRequest = req;
                    
                    var NetSocket = new net.Socket();
                    NetSocket.setKeepAlive(true,5000);
                    
                    NetSocket.on('data', onData);
                    NetSocket.connect(8888, self.config.IP_ADDR, (function(req){
                        console.log('connection : OK')
                        self.netSocket.isEnded = false;
                        NetSocket.write(req);
                    }).bind(null, req));
                    NetSocket.on('error', function(err){ 
                        console.log('err. : ', err);
                        self.paymentStatus[self.config.TOKEN] = 'error'; 
                    });
                    NetSocket.on('end', function(e){ 
                        console.log('Payment: Socket Close ... Can t reconnect', e);
                        self.netSocket.isEnded = true;
                    });
                    self.netSocket = NetSocket
                    return new Promise(function(resolve){
                        return NetSocket.on("close", resolve)
                    });
            },
            
            onData = function(line){
                console.log('PAYMENT : GET DATA ...');
                if (self.netSocket && self.netSocket.isEnded === false){
                   self.netSocket.end();
                } else {
                    try {
                      if (self.serialPort.isClosed === false)
                            self.serialPort.close()
                    } catch(e){};
                }



                self.config.lastReceive = line.toString();
                let tempKey = line.toString().charAt(0);
                if (typeof self.config.listAction[tempKey] == 'function')
                    self.config.listAction[tempKey](self.config);
                else if (self.config.listAction['default'])
                    self.config.listAction['default'](self.config);
                else
                    console.log(self.config.listAction, line);
            },
            /*
            *
            *  TPEBySerialPort : init Serial Communication
            *  Take in params the value to send on TPE
            *  & an "Array" (Object) with the expected value binded with their function to run
            */
            TPEBySerialPort = function(req, listAction){
                const port = new SerialPort(self.config.DEVICE, {
                    baudRate:self.config.DEVICE_RATE,
		            parser: new SerialPort.parsers.Readline()
                });
                self.config.listAction = listAction;
                self.serialPort = port;
                port.on('data', onData);
                port.on('open',(function(req){
                    port.write(req);
                    self.serialPort.isClosed = false;
                }).bind(null, req));
                port.on('error', function(err){ 
                    console.log('err. : ', err);
                    self.paymentStatus[self.config.TOKEN] = 'error'; 
                });
                port.on('close', function(err){ 
                    console.log('close : ', err) 
                    self.serialPort.isClosed = true;
                });
                return new Promise(function(resolve){
                    return port.on("close", resolve)
                });
            },
                /*
                *
                *    generate_lrc : used for the signature of message
                *
                *
                */
                generate_lrc = function(real_msg_with_etx){
                    self.paymentStatus[self.config.TOKEN] = 'progress';
                    let lrc = 0, text = real_msg_with_etx.split('');
                    for (let i in text){
                        lrc ^= text[i].charCodeAt(0);
                    }
                    console.log('lrc => ', lrc);
                    return lrc;
                },  
                send_eot_signal = function(){
                    self.paymentStatus[self.config.TOKEN] = 'progress';
                    console.log('Signal EOT sent to terminal');
                    let listAction = {};
                    listAction[String.fromCharCode(5)] = function(){
                        console.log('ENQ received from terminal');
                        send_enq_signal(self.config);
                        }; 
                   return (self.config.IP_ADDR ? TPEByEthernet : TPEBySerialPort )(String.fromCharCode(4),listAction); //EOT
                },
                send_enq_signal = function(){
                    console.log('Signal ACK sent to terminal');
                    self.paymentStatus[self.config.TOKEN] = 'progress';
                    let listAction = {};
                    listAction['default'] = get_serial_answer_from_terminal;
                    return (self.config.IP_ADDR ? TPEByEthernet : TPEBySerialPort )(String.fromCharCode(6),listAction); //ACK
                },

                 /*
                *
                *  get_serial_socket_from_terminal : Get the confirmation of the Payment
                *  (or cancelation)
                *
                *
                */
               get_socket_answer_from_terminal = function(){
                let response = {};
                console.log('Now expecting answer from Terminal');
                self.config.lastReceive = self.config.lastReceive.
                                        substr(1).substr(0, self.config.lastReceive.length-3);
                console.log(self.config.lastReceive);
                response = {
                    'pos_number'        : self.config.lastReceive.charAt(0),
                    /* 
                    * 0 si transaction acceptée
                    * 1 NOK : auth
                    * 2 phonie
                    * 3 force
                    * 4 refused
                    * 5 carte interdite
                    * 6 Erreur : cancelled by cash registe
                    * 7 si transaction non aboutie
                    * 8 impossible
                    * 9 si requête prise en compte (Unknown response)
                    */
                    'transaction_result': self.config.lastReceive.charAt(1),
                    'amount_msg'        : self.config.lastReceive.substr(0, 8),
                    /* 
                    * Carte bancaire   1
                    * Chèque           C
                    * Amex             2
                    * CETELEM          3
                    * Cofinoga         5
                    * Diners           6
                    * Franfinance      8
                    * JCB              9
                    * Accord-finances  A
                    * Moneo            O
                    * CUP              U
                    * FINTRAX EMV      F
                    * autres           0
                    */
                    'payment_mode'      : self.config.lastReceive.charAt(10)
                };
                self.paymentStatus[self.config.TOKEN] = response;
                console.log('response : ', JSON.stringify(response));
            },


                /*
                *
                *  get_serial_answer_from_terminal : Get the confirmation of the Payment
                *  (or cancelation)
                *
                *
                */
                get_serial_answer_from_terminal = function(){
                    let response = {};
                    console.log('Now expecting answer from Terminal');
                    self.config.lastReceive = self.config.lastReceive.
                                            substr(1).substr(0, self.config.lastReceive.length-3);
                    console.log(self.config.lastReceive);
                    response = {
                        'pos_number'        : self.config.lastReceive.substr(0, 2),
                        /* 
                        * 0 si transaction acceptée
                        * 7 si transaction non aboutie
                        * 9 si requête prise en compte
                        */
                        'transaction_result': self.config.lastReceive.charAt(2),
                        'amount_msg'        : self.config.lastReceive.substr(3, 8),
                        /* 
                        * Carte bancaire   1
                        * Chèque           C
                        * Amex             2
                        * CETELEM          3
                        * Cofinoga         5
                        * Diners           6
                        * Franfinance      8
                        * JCB              9
                        * Accord-finances  A
                        * Moneo            O
                        * CUP              U
                        * FINTRAX EMV      F
                        * autres           0
                        */
                        'payment_mode'      : self.config.lastReceive.charAt(11),
                        'currency_numeric'  : self.config.lastReceive.substr(12, 3),
                        'private'           : self.config.lastReceive.substr(15, 11)
                    };
                    self.paymentStatus[self.config.TOKEN] = response;
                    console.log('response : ', JSON.stringify(response));
                },

                /*
                *
                *  Prepare & Send Payment Request to TPE
                *  !!!! DON'T CHANGE THE ORDER !!!
                *
                */
                send_serial_message    = function(){
                    console.log('Send Message  ...');
                    let data = {
                        'pos_number': '01',
                        'amount_msg': ('0000000' + (self.config.AMOUNT * 100)).substr(-8),
                        'answer_flag': '0',
                        'payment_mode': self.config.PAYMENT_MODE  == 'check' ? 'C' : '1', 
                        'transaction_type': '0',
                        'currency_numeric': 978, // EUR => https://fr.iban.com/currency-codes
                        'private': '          ',
                        'delay': 'A010',
                        'auto': 'B010'
                    };
                    let msg = Object.keys(data).map( k => data[k] ).join('');
                    if (msg.length > 34) return console.log('ERR. : failed data > 34 characters.', msg);
                    let real_msg_with_etx = msg.concat(String.fromCharCode(3));//ETX
                    let lrc = generate_lrc(real_msg_with_etx);
                    //STX + msg + lrc
                    let tpe_msg = (String.fromCharCode(2)).concat(real_msg_with_etx).concat(String.fromCharCode(lrc));
                    console.log('Real message to send =', msg);
                    console.log('Message sent to terminal');
                    let  listAction = {};
                    listAction[String.fromCharCode(6)] = ()=>{ 
                        console.log('ACK received from terminal'); 
                        send_eot_signal();
                    };
                    return (self.config.IP_ADDR ? TPEByEthernet : TPEBySerialPort )(tpe_msg, listAction);
                },
                /*
                *
                *  Prepare & Send Payment Request to TPE
                *  !!!! DON'T CHANGE THE ORDER !!!
                *
                */
               send_socket_message    = function(){
                console.log('Send Message  ...');
                let data = {
                    'pos_number'        : '1',
                    'amount_msg'        : ('0000000' + (self.config.AMOUNT * 100)).substr(-8),
                    'answer_flag'       : '0',
                    'payment_mode'      : self.config.PAYMENT_MODE  == 'check' ? 'C' : '1', 
                    'transaction_type'  : '0',
                    'currency_numeric'  : 978, // EUR => https://fr.iban.com/currency-codes
                    'private'           : '          ',
                    'delay'             : '0',
                    'auto'              : '0'
                };
                let msg = Object.keys(data).map( k => data[k] ).join('');
                if (msg.length > 34) return console.log('ERR. : failed data > 34 characters.', msg);
                let real_msg_with_etx = msg.concat(String.fromCharCode(3));//ETX
                let lrc = generate_lrc(real_msg_with_etx);
                //STX + msg + lrc
                let tpe_msg = (String.fromCharCode(2)).concat(real_msg_with_etx).concat(String.fromCharCode(lrc));
                console.log('Real message to send =', msg);
                console.log('Message sent to terminal');
                let  listAction = {};
                listAction['default'] = get_socket_answer_from_terminal;
                return (self.config.IP_ADDR ? TPEByEthernet : TPEBySerialPort )(tpe_msg, listAction);
            },
                /*
                *
                *  Init the TPE
                *
                *
                */
                transaction_start = function(){
                    try {
                        // If Payment by Network directly send payment !
                        if (self.config.IP_ADDR) return send_socket_message();
                        // Init
                        console.log('Signal ENQ sent to terminal');
                        let listAction = {};
                        listAction[String.fromCharCode(6)] = function(){    //ACK
                            console.log('ACK received from terminal');
                            send_serial_message();
                        };
                        listAction[String.fromCharCode(5)] = function(){    //ENCK
                            if (self.config.MAX_RETRY <= 0) {
                                console.log('Payment : Unable to make the ACK > ENQ workflow ... ')
                                console.log('try to directly make payment');
                                send_serial_message();
                            }
                            console.log('ENQ received from terminal ... Retry');
                            let listAction = {};
                            self.config.MAX_RETRY -= 1;
                            listAction[String.fromCharCode(6)] = transaction_start;
                            return (self.config.IP_ADDR ? TPEByEthernet : TPEBySerialPort )
                                    (String.fromCharCode(6), listAction);
                        };
                        return (self.config.IP_ADDR ? TPEByEthernet : TPEBySerialPort )
                                (String.fromCharCode(5), listAction);
                        } catch (e){
                        return console.log(e);            
                        }
                };

            /*
            *
            *  Main - Poc 
            *
            *
            *
            */
            self.config = {
                TOKEN           : typeof postParam.token        == 'string'  ? postParam.token : (new Date()).getTime(),
                DEVICE          : typeof postParam.device       == 'string'    && postParam.device.length >= 3 ? postParam.device    :'/dev/ttyACM0',
                IP_ADDR         : typeof postParam.networkIp    == 'string'    && postParam.networkIp.length > 7 ? postParam.networkIp : false,
                DEVICE_RATE     : 9600,
                MAX_RETRY       : 3,
                PAYMENT_MODE    : 'card',
                AMOUNT          : typeof postParam.amount == 'string' && 
                        postParam.amount.match(/[0-9]{1,9}\.[0-9]{1,9}/) ? parseFloat(postParam.amount) : 0.01
            };
            console.log(postParam, self.config);
            return resolve(transaction_start());
        });    
    }
        
}


/*#############################################################################
 *
 *
 *      Init the Web Server to monitor TPE
 *	token is random value to identify a payment 
 *
 *
 *
 *
 *#############################################################################*/
const   express = require('express'), app = express(), 
        bodyParser = require('body-parser');
 var paymentClass = new PaymentService;
 app.listen(3000, function () { console.log('TPE WebServer started !') });
 app.use( bodyParser.urlencoded({ extended: true  }));
 app.use(bodyParser.json())
/*
Example with Network :
x= new XMLHttpRequest();
x.open('POST', 'http://127.0.0.1:3000/api/payment/tpe');
x.addEventListener('load', e=> console.log(e.target.response));
x.send('token=123&device=&networkIp=127.0.0.1&amount=3.0');

Exammple with USB  / COM:
x= new XMLHttpRequest();
x.open('POST', 'http://127.0.0.1:3000/api/payment/tpe');
x.addEventListener('load', e=> console.log(e.target.response));
x.send('token=123&device=COM2&networkIp=&amount=3.0');

*/
app.get('/', function(req, res){ res.send('wait instruction');});
 app.post('/api/payment/tpe', function (req, res) {
	 paymentClass.makePayment(req.body);
 });

/*

Example to get information from the token 123
x= new XMLHttpRequest();
x.open('POST', 'http://127.0.0.1:3000/api/payment/tpeInfo');
x.addEventListener('load', e=> console.log(e.target.response));
x.send('token=123');

*/
 app.post('/api/payment/tpeInfo', function (req, res) {
  	res.send(paymentClass.getPaymentState(req.body));
 });
